'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

class Detail {
    static Unescape(str) {
        let res = "";
        for(let i = 0; i < str.length; i++){
            let char = str[i];
            if (char == "\\" && i != str.length - 1) {
                char = str[i + 1];
                switch(char){
                    case "n":
                    case "\n":
                        res += "\n";
                        break;
                    case "r":
                        res += "\r";
                        break;
                    case "'":
                    case "\"":
                    case "\\":
                        res += char;
                        break;
                    default:
                        res += "\\";
                        res += char;
                        break;
                }
                continue;
            }
            res += char;
        }
        return res.trim();
    }
    static Escape(str) {
        let res = "";
        for(let i = 0; i < str.length; i++){
            const char = str[i];
            switch(char){
                case "\n":
                    res += "\\n";
                    break;
                case "\r":
                    res += "\\r";
                    break;
                case "'":
                case "\"":
                case "\\":
                    res += "\\";
                    res += char;
                    break;
                default:
                    res += char;
                    break;
            }
        }
        return res;
    }
}

class Emitter {
    containsSpecials(value) {
        return /[:,'"\[\]\{\}]/gm.test(value);
    }
    emitConfigValue(value, indent = 0, isLast = true, commas = true, apostrophes = true) {
        const _indent = " ".repeat(indent * 2);
        if (value instanceof Array) {
            //os.write('[\n');
            this.stream += "[\n";
            for(let i = 0; i < value.length; i++){
                //os.write(_indent);
                this.stream += _indent;
                this.emitConfigValue(value[i], indent + 1, i == value.length - 1, commas, apostrophes);
            }
            //os.write(_indent.repeat(indent - 1) + `${isLast || !commas ? ']\n' : '],\n'}`);
            this.stream += _indent.repeat(indent - 1) + `${isLast || !commas ? "]\n" : "],\n"}`;
        } else if (value instanceof Object) {
            if (indent > 0) //os.write('{\n');
            this.stream += "{\n";
            const keys = Object.keys(value);
            for(let i = 0; i < keys.length; i++){
                const key = keys[i];
                const _value = value[key];
                if (_value == null) continue;
                //os.write(_indent + key + ':');
                this.stream += _indent + key + ": ";
                this.emitConfigValue(_value, indent + 1, i == keys.length - 1, commas, apostrophes);
            }
            if (indent > 0) //os.write(_indent.repeat(indent - 1) + `${isLast || !commas ? '}\n' : '},\n'}`);
            this.stream += " ".repeat((indent - 1) * 2) + `${isLast || !commas ? "}\n" : "},\n"}`;
        } else {
            let escaped = null;
            if (typeof value === "boolean") {
                escaped = Detail.Escape(String(value));
            } else if (typeof value === "string") {
                escaped = Detail.Escape(value);
            } else if (typeof value === "number") {
                escaped = Detail.Escape(value.toString());
            }
            if (escaped === null) {
                throw new Error(`[CFG-READER] can not emit value of type: ${typeof value}. (you passed an invalid data type)`);
            }
            const useApostrophes = apostrophes || this.containsSpecials(escaped);
            this.stream += (useApostrophes ? "'" : "") + escaped + (useApostrophes ? "'" : "") + (commas ? "," : "") + "\n";
        }
    }
    constructor(){
        this.stream = "";
    }
}

var TokenType;
(function(TokenType) {
    TokenType[TokenType["ArrayStart"] = 0] = "ArrayStart";
    TokenType[TokenType["ArrayEnd"] = 1] = "ArrayEnd";
    TokenType[TokenType["DictStart"] = 2] = "DictStart";
    TokenType[TokenType["DictEnd"] = 3] = "DictEnd";
    TokenType[TokenType["Key"] = 4] = "Key";
    TokenType[TokenType["Scalar"] = 5] = "Scalar";
})(TokenType || (TokenType = {}));
var ErrorType;
(function(ErrorType) {
    ErrorType[ErrorType["KeyExpected"] = 0] = "KeyExpected";
    ErrorType[ErrorType["InvalidToken"] = 1] = "InvalidToken";
    ErrorType[ErrorType["UnexpectedEOF"] = 2] = "UnexpectedEOF";
})(ErrorType || (ErrorType = {}));
function isInt(value) {
    return /^-?\d+$/.test(value);
}
function isFloat(value) {
    const x = value.split(".");
    return x.length == 2 && x.every(isInt);
}
let Token = class Token {
    constructor(_type, _value = "", _pos = 0, _line = 0, _col = 0){
        this.type = _type;
        this.value = _value;
        this.pos = _pos;
        this.line = _line;
        this.col = _col;
    }
};
class Parser {
    parse() {
        this.tokenize();
        return this.parseToken();
    }
    unread() {
        return this.buffer.length - this.readPos;
    }
    peek(offset = 0) {
        const idx = this.readPos + offset;
        return this.buffer[idx];
    }
    get() {
        this.column++;
        if (this.peek() == "\n") {
            this.line++;
            this.column = 0;
        }
        //const currPos = this.readPos;
        //this.readPos++;
        //return this.buffer[currPos];
        return this.buffer[this.readPos++];
    }
    skip(n = 1) {
        for(let i = 0; i < n; i++){
            this.column++;
            if (this.peek(i) == "\n") {
                this.line++;
                this.column = 0;
            }
        }
        this.readPos += n;
    }
    skipNextToken() {
        while(this.unread() > 0){
            if (this.peek() == " " || this.peek() == "\n" || this.peek() == "\r" || this.peek() == "\t" || this.peek() == ",") {
                this.skip();
            } else if (this.peek() == "#") {
                this.skip();
                while(this.unread() > 0 && this.peek() != "\n" && this.peek() != "#"){
                    this.skip();
                }
                if (this.unread() > 0) {
                    this.skip();
                }
            } else {
                break;
            }
        }
    }
    tokenize() {
        this.tokens.push(new Token(TokenType.DictStart));
        while(this.unread() > 0){
            this.skipNextToken();
            if (this.unread() == 0) {
                break;
            }
            if (this.peek() == "[") {
                this.skip();
                this.tokens.push(new Token(TokenType.ArrayStart, "", this.readPos, this.line, this.column));
            } else if (this.peek() == "]") {
                this.skip();
                this.tokens.push(new Token(TokenType.ArrayEnd, "", this.readPos, this.line, this.column));
            } else if (this.peek() == "{") {
                this.skip();
                this.tokens.push(new Token(TokenType.DictStart, "", this.readPos, this.line, this.column));
            } else if (this.peek() == "}") {
                this.skip();
                this.tokens.push(new Token(TokenType.DictEnd, "", this.readPos, this.line, this.column));
            } else {
                let val = "";
                if (this.peek() == "'" || this.peek() == '"') {
                    const start = this.get();
                    if (this.peek() != start) {
                        while(this.unread() > 1 && (this.peek() == "\\" || this.peek(1) != start)){
                            if (this.peek() == "\n" || this.peek() == "\r") {
                                if (this.get() == "\r" || this.peek() == "\n") {
                                    this.skip();
                                }
                                val += "\n";
                                continue;
                            }
                            val += this.get();
                        }
                        if (this.unread() > 0) {
                            val += this.get();
                        }
                        if (this.unread() == 0) {
                            throw new Error(this.createParseError(ErrorType.UnexpectedEOF, this.line, this.column));
                        }
                    }
                    this.skip();
                } else {
                    while(this.unread() > 0 && this.peek() != "\n" && this.peek() != ":" && this.peek() != "," && this.peek() != "]" && this.peek() != "}" && this.peek() != "#"){
                        val += this.get();
                    }
                }
                val = Detail.Unescape(val);
                if (this.unread() > 0 && this.peek() == ":") {
                    this.tokens.push(new Token(TokenType.Key, val, this.readPos, this.line, this.column));
                } else {
                    this.tokens.push(new Token(TokenType.Scalar, val, this.readPos, this.line, this.column));
                }
                if (this.unread() > 0 && (this.peek() == ":" || this.peek() == ",")) {
                    this.skip();
                }
            }
        }
        this.tokens.push(new Token(TokenType.DictEnd));
        return; // end
    }
    createParseError(type, token, col) {
        let line;
        if (token instanceof Token) {
            col = token.col;
            line = token.line;
        } else line = token;
        line++;
        if (col) col++;
        const place = /*this.filePath
      ? `${this.filePath}:${line}:${col}`
      : */ `${line}:${col}`;
        const base = `[CFG-READER] error at line ${place} -> `;
        switch(type){
            case ErrorType.KeyExpected:
                return base + `key expected`;
            case ErrorType.InvalidToken:
                return base + `invalid token`;
            case ErrorType.UnexpectedEOF:
                return base + `unexpected end of file`;
        }
    }
    parseToken() {
        const token = this.tokens[this.tokIdx];
        switch(token.type){
            case TokenType.Scalar:
                // FIXME:
                if (token.value === "true" || token.value === "false" || token.value === "yes" || token.value === "no") {
                    return token.value === "true" || token.value === "yes";
                } else if (isInt(token.value) || isFloat(token.value)) {
                    return parseFloat(token.value);
                } else {
                    return token.value;
                }
            case TokenType.ArrayStart:
                const list = [];
                while(this.tokIdx < this.tokens.length - 1 && this.tokens[this.tokIdx + 1].type != TokenType.ArrayEnd){
                    this.tokIdx++;
                    list.push(this.parseToken());
                }
                this.tokIdx++;
                return list;
            case TokenType.DictStart:
                const dict = {};
                while(this.tokIdx < this.tokens.length - 1 && this.tokens[this.tokIdx + 1].type != TokenType.DictEnd){
                    this.tokIdx++;
                    const nextTok = this.tokens[this.tokIdx];
                    if (nextTok.type != TokenType.Key) {
                        throw new Error(this.createParseError(ErrorType.KeyExpected, nextTok));
                    }
                    const key = nextTok.value;
                    this.tokIdx++;
                    dict[key] = this.parseToken();
                }
                this.tokIdx++;
                return dict;
        }
        throw new Error(this.createParseError(ErrorType.InvalidToken, token));
    }
    // public filePath: string;
    constructor(content){
        this.tokens = [];
        this.readPos = 0;
        this.line = 0;
        this.column = 0;
        this.tokIdx = 0;
        this.buffer = content;
    }
}

function parse(content) {
    const parser = new Parser(content);
    return parser.parse();
}
function serialize(config, useCommas = false, useApostrophe = false) {
    const emitter = new Emitter();
    emitter.emitConfigValue(config, 0, true, useCommas, useApostrophe);
    return emitter.stream;
}

exports.parse = parse;
exports.serialize = serialize;
